import random

APIKEY = "lstzFDEOhfFNMLikKa0am9mgEKLBl49T"

ASSET_TYPE = ["etf", "fund", "stock"]

EXCHANGE = {
    'ARCX': 'NYSE Arca (Pacific Stock Exchange)',
    'BATS': 'BATS Exchange inc',
    'CHIA': 'Chi-X Australia',
    'E0WWE$$ALL': 'WORLDWIDE_EQUITY',
    'FINR': 'OTC Composite',
    'IPSX': 'equity_exchange_IPSX',
    'IXUS': "Standard + Poor's Indices",
    'MABX': 'Madrid stock exchange Equities',
    'MSCO': 'MSCI (EOD Only)',
    'MSTARFund': 'Morningstar Fund',
    'OTCM': 'OTC Markets (was Pinksheets)',
    'USCO': 'US composite market',
    'XAMS': 'NYSE Euronext Amsterdam Equities',
    'XASE': 'American Stock Exchange',
    'XASX': 'Australia Stock Exchange',
    'XATH': 'Athens Stock Exchange',
    'XBER': 'Berlin',
    'XBKK': 'Thailand Stock Exchange',
    'XBOM': 'Mumbai Stock Exchange',
    'XBRU': 'NYSE Euronext Brussels Equities',
    'XCNQ': 'CNSX (Canada National Stock Exchange)',
    'XCSE': 'OMX exchange Copenhagen',
    'XDUB': 'Irish Stock Exchange',
    'XDUS': 'Düsseldorf',
    'XETR': 'XETRA Frankfurt',
    'XEUR': 'EUREX (was DTB Deutsche Termin Börse)',
    'XFRA': 'Frankfurter Börse (or Germany in general)',
    'XHAM': 'Hamburg',
    'XHAN': 'Hannover (via CEF)',
    'XHEL': 'Helsinki (Equities)',
    'XHKF': 'Hong Kong Stock Exchange / Hong Kong Futures Exchange / Hang Seng Indices',
    'XHKG': 'HONG_KONG',
    'XICE': 'OMX Exchange Iceland',
    'XIST': 'Istanbul Stock Exchange',
    'XKOS': 'Korea Stock Exchange (KOSCOM)',
    'XLIS': 'NYSE Euronext Lisbon',
    'XLIT': 'OMX Exchange Vilnius',
    'XLON': 'LSE (SEAQ national)',
    'XLUX': 'Luxembourg Stock Exchange',
    'XMEX': 'Bolsa de valores Mexico',
    'XMIL': 'Milan Stock Exchange (Borsa Italia)',
    'XMUN': 'Munchen',
    'XNAS': 'NASDAQ national market',
    'XNSE': 'India National Stock Exchange',
    'XNYS': 'New York Stock Exchange',
    'XNZE': 'New Zealand Stock Exchange',
    'XOSE': 'Osaka',
    'XOSL': 'Oslo bors',
    'XOTC': 'OTCBB, over the counter bulletin board equities',
    'XPAR': 'NYSE Euronext Paris Equities',
    'XRIS': 'OMX Exchange Riga',
    'XSES': 'Singapore Stock Exchange',
    'XSHE': 'Shenzhen Stock Exchange',
    'XSHG': 'Shanghai Stock Exchange',
    'XSTO': 'NASDAQ OMX Stockholm',
    'XSTU': 'Stuttgart',
    'XSWX': 'SIX Feed (was SWX)',
    'XTAI': 'Taiwan Stock Exchange',
    'XTAL': 'Tallin Stock Exchange',
    'XTKS': 'Tokyo Stock Exchange',
    'XTSE': 'Toronto Stock Exchange',
    'XWAR': 'Warsaw Stock Exchange',
    'XWBO': 'Wiener Boerse',


}


FILTER_TYPE = [
        'basics',
        'dividends',
        'fees-and-expenses',
        'keyword',
        'other',
        'performance',
        'profitability-and-growth',
        'purchase-requirements',
        'ratings',
        'risk',
        'valuation',
 ]

LANGUAGE = {'en-ea': 'Asia',
            'nl': 'Benelux',
            'en-ca': 'CanadaEnglish',
            'fr-ca': 'CanadaFrench',
            'en-eu': 'CentralEurope',
            'fr': 'France',
            'de': 'Germany',
            'it': 'Italy',
            'ja': 'Japan',
            'en-nd': 'Nordics',
            'es': 'Spain',
            'sv': 'Sweden',
            'en-gb': 'UnitedKingdom'}

USER_AGENTS = [
    "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10.5; en-US; rv:1.9.1b3) Gecko/20090305 Firefox/3.1b3 GTB5",
    "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10.5; ko; rv:1.9.1b2) Gecko/20081201 Firefox/3.1b2",
    "Mozilla/5.0 (X11; U; SunOS sun4u; en-US; rv:1.9b5) Gecko/2008032620 Firefox/3.0b5",
    "Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.8.1.12) Gecko/20080214 Firefox/2.0.0.12",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; cs; rv:1.9.0.8) Gecko/2009032609 Firefox/3.0.8",
    "Mozilla/5.0 (X11; U; OpenBSD i386; en-US; rv:1.8.0.5) Gecko/20060819 Firefox/1.5.0.5",
    "Mozilla/5.0 (Windows; U; Windows NT 5.0; es-ES; rv:1.8.0.3) Gecko/20060426 Firefox/1.5.0.3",
    "Mozilla/5.0 (Windows; U; WinNT4.0; en-US; rv:1.7.9) Gecko/20050711 Firefox/1.0.6",
    "Mozilla/5.0 (Windows; Windows NT 6.1; rv:2.0b2) Gecko/20100720 Firefox/4.0b2",
    "Mozilla/5.0 (X11; Linux x86_64; rv:2.0b4) Gecko/20100818 Firefox/4.0b4",
    "Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.2) Gecko/20100308 Ubuntu/10.04 (lucid) Firefox/3.6 GTB7.1",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:2.0b7) Gecko/20101111 Firefox/4.0b7",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:2.0b8pre) Gecko/20101114 Firefox/4.0b8pre",
    "Mozilla/5.0 (X11; Linux x86_64; rv:2.0b9pre) Gecko/20110111 Firefox/4.0b9pre",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:2.0b9pre) Gecko/20101228 Firefox/4.0b9pre",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:2.2a1pre) Gecko/20110324 Firefox/4.2a1pre",
    "Mozilla/5.0 (X11; U; Linux amd64; rv:5.0) Gecko/20100101 Firefox/5.0 (Debian)",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:6.0a2) Gecko/20110613 Firefox/6.0a2",
    "Mozilla/5.0 (X11; Linux i686 on x86_64; rv:12.0) Gecko/20100101 Firefox/12.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:15.0) Gecko/20120716 Firefox/15.0a2",
    "Mozilla/5.0 (X11; Ubuntu; Linux armv7l; rv:17.0) Gecko/20100101 Firefox/17.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:21.0) Gecko/20130328 Firefox/21.0",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64; rv:22.0) Gecko/20130328 Firefox/22.0",
    "Mozilla/5.0 (Windows NT 5.1; rv:25.0) Gecko/20100101 Firefox/25.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.8; rv:25.0) Gecko/20100101 Firefox/25.0",
    "Mozilla/5.0 (Windows NT 6.1; rv:28.0) Gecko/20100101 Firefox/28.0",
    "Mozilla/5.0 (X11; Linux i686; rv:30.0) Gecko/20100101 Firefox/30.0",
    "Mozilla/5.0 (Windows NT 5.1; rv:31.0) Gecko/20100101 Firefox/31.0",
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:33.0) Gecko/20100101 Firefox/33.0",
    "Mozilla/5.0 (Windows NT 10.0; WOW64; rv:40.0) Gecko/20100101 Firefox/40.0",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.12; rv:58.0) Gecko/20100101 Firefox/58.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/525.19 (KHTML, like Gecko) Chrome/1.0.154.53 Safari/525.19",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/525.19 (KHTML, like Gecko) Chrome/1.0.154.36 Safari/525.19",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.10 (KHTML, like Gecko) Chrome/7.0.540.0 Safari/534.10",
    "Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US) AppleWebKit/534.4 (KHTML, like Gecko) Chrome/6.0.481.0 Safari/534.4",
    "Mozilla/5.0 (Macintosh; U; Intel Mac OS X; en-US) AppleWebKit/533.4 (KHTML, like Gecko) Chrome/5.0.375.86 Safari/533.4",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/532.2 (KHTML, like Gecko) Chrome/4.0.223.3 Safari/532.2",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/532.0 (KHTML, like Gecko) Chrome/4.0.201.1 Safari/532.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US) AppleWebKit/532.0 (KHTML, like Gecko) Chrome/3.0.195.27 Safari/532.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/530.5 (KHTML, like Gecko) Chrome/2.0.173.1 Safari/530.5",
    "Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US) AppleWebKit/534.10 (KHTML, like Gecko) Chrome/8.0.558.0 Safari/534.10",
    "Mozilla/5.0 (X11; U; Linux x86_64; en-US) AppleWebKit/540.0 (KHTML,like Gecko) Chrome/9.1.0.0 Safari/540.0",
    "Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/534.14 (KHTML, like Gecko) Chrome/9.0.600.0 Safari/534.14",
    "Mozilla/5.0 (X11; U; Windows NT 6; en-US) AppleWebKit/534.12 (KHTML, like Gecko) Chrome/9.0.587.0 Safari/534.12",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.13 (KHTML, like Gecko) Chrome/9.0.597.0 Safari/534.13",
    "Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US) AppleWebKit/534.16 (KHTML, like Gecko) Chrome/10.0.648.11 Safari/534.16",
    "Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US) AppleWebKit/534.20 (KHTML, like Gecko) Chrome/11.0.672.2 Safari/534.20",
    "Mozilla/5.0 (Windows NT 6.0) AppleWebKit/535.1 (KHTML, like Gecko) Chrome/14.0.792.0 Safari/535.1",
    "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/535.2 (KHTML, like Gecko) Chrome/15.0.872.0 Safari/535.2",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/535.7 (KHTML, like Gecko) Chrome/16.0.912.36 Safari/535.7",
    "Mozilla/5.0 (Windows NT 6.0; WOW64) AppleWebKit/535.11 (KHTML, like Gecko) Chrome/17.0.963.66 Safari/535.11",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_6_8) AppleWebKit/535.19 (KHTML, like Gecko) Chrome/18.0.1025.45 Safari/535.19",
    "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/535.24 (KHTML, like Gecko) Chrome/19.0.1055.1 Safari/535.24",
    "Mozilla/5.0 (Windows NT 6.2) AppleWebKit/536.6 (KHTML, like Gecko) Chrome/20.0.1090.0 Safari/536.6",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.1 (KHTML, like Gecko) Chrome/22.0.1207.1 Safari/537.1",
    "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.15 (KHTML, like Gecko) Chrome/24.0.1295.0 Safari/537.15",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/27.0.1453.93 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/28.0.1467.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/30.0.1599.101 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/31.0.1623.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/34.0.1847.116 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/37.0.2062.103 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2214.38 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2490.71 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.103 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.62 Safari/537.36",
]


def random_user_agent() -> str:
    """
    This function selects a random User-Agent from the User-Agent list, . User-Agents are used in
    order to avoid the limitations of the requests to morningstar.com. The User-Agent is
    specified on the headers of the requests and is different for every request.



    Returns:
       :obj:`str` - user_agent:
            The returned:obj:`str` is the name of a random User-Agent, which will be passed on the
            headers of a request so to avoid restrictions due to the use of multiple requests from the
            same User-Agent.

    """

    return random.choice(USER_AGENTS)



from .security import Security
import datetime

class Stock(Security):
    """
    Main class to access data about stocks, inherit from Security class
    Args:
        term (str): text to find a stock, can be a name, part of a name or the isin of the stocks
        language (str): language of the data, default is "en-gb"
        filters (dict) : filter, use the method search_filter() to find the different possible filter keys
        itemRange (int) : index of stocks to return (must be inferior to PageSize)
        pageSize (int): number of securities to return
        page (int): page to return
        sortby (str) : sort by a field
        ascending (bool) : True sort by ascending order, False sort by descending order
        proxies (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}

    Examples:
        >>> Stock("ab",page=5, pageSize=5,itemRange=0,sortby="name", ascending=False)
        >>> Stock('US0378331005')

    Raises:
        TypeError: raised whenever the parameter type is not the type expected
        ValueError : raised whenever the parameter is not valid or no stock found

    """

    def __init__(
        self,
        term:str,
        language:str="en-gb",
        filters:dict=None,
        itemRange:int=0,
        pageSize:int=10,
        page:int=1,
        sortby:str=None,
        ascending:bool=True,
        proxies:dict=None,
    ) -> None:
        
        stock_filter = {"investmentType" : 'EQ'}
        if filters:
            stock_filter = stock_filter | filters

        super().__init__(
            term=term,
            asset_type="stock",
            language=language,
            filters=stock_filter,
            itemRange=itemRange,
            pageSize=pageSize,
            page=page,
            sortby=sortby,
            ascending=ascending,
            proxies=proxies,
        )

    def analysisData(self) -> dict:
        """
        This function retrieves general data about the stock.

        Returns:
            dict with general data about stock

        Examples:
            >>> Stock("US0378331005").analysisData()

        """
        return self.GetData("morningstarTake/v3", url_suffix="analysisData").json()

    def analysisReport(self) -> dict:
        """
        This function retrieves the analysis of the stock.

        Returns:
            dict with analyst overview

        Examples:
            >>> Stock("US0378331005").analysisReport()

        """
        return self.GetData("morningstarTake/v4", url_suffix="analysisReport").json()

    def balanceSheet(self, 
                     period:str="annual",
                     reportType:str="original",
                     export:bool=False,
                     folderPath:str="."
                     ) -> dict:
        """
        This function retrieves the balance sheet.

        Args:
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated
            export (bool) : if True, the file is downloaded in the folderPath
            folderPath (str) : path to save the file if export is True

        Returns:
            dict with balance sheet

        Examples:
            >>> Stock("US0378331005").balanceSheet('quarterly', 'original')
            >>> Stock("US0378331005").balanceSheet('annual', 'restated')

        """

        return self.financialStatement(
            "balancesheet", 
            period=period, 
            reportType=reportType,
            export=export,
            folderPath=folderPath
        )

    def boardOfDirectors(self) -> dict:
        """
        This function retrieves information about the board of directors.

        Returns:
            dict with board of directors information

        Examples:
            >>> Stock("Alphabet Inc Class A").boardOfDirectors()

        """
        return self.GetData("insiders/boardOfDirectors").json()

    def cashFlow(self, 
                 period:str="annual",
                 reportType:str="original",
                 export:bool=False,
                 folderPath:str="."
                 ) -> dict:
        """
        This function retrieves the cash flow.

        Args:
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated
            export (bool) : if True, the file is downloaded in the folderPath
            folderPath (str) : path to save the file if export is True

        Returns:
            dict with cash flow

        Examples:
            >>> Stock("US0378331005").cashFlow('annual', 'restated')
            >>> Stock("US0378331005").cashFlow('quarterly', 'restated')

        """

        return self.financialStatement("cashflow", 
                                       period=period, 
                                       reportType=reportType,
                                       export=export,
                                       folderPath=folderPath)


    def companyProfile(self) -> dict:
        """
        This function retrieves the company profile

        Returns:
            dict with company information

        Examples:
            >>> Stock("US0378331005").companyProfile()

        """
        return self.GetData("companyProfile", url_suffix="").json()

    def dividends(self) -> dict:
        """
        This function retrieves the dividends of the stock.

        Returns:
            dict with dividends

        Examples:
            >>> Stock("US0378331005").dividends()

        """
        return self.GetData("dividends/v4").json()

    def esgRisk(self) -> dict:
        """
        This function retrieves the esg risk of the stock.

        Returns:
            dict with esg risk

        Examples:
            >>> Stock("US0378331005").esgRisk()

        """
        return self.GetData("esgRisk").json()
    
    def financialHealth(self) -> dict:
        """
        This function retrieves the financial health of the stock.

        Returns:
            dict with financial health

        Examples:
            >>> Stock("US0378331005").financialHealth()

        """
        return self.GetData("keyMetrics/financialHealth", url_suffix="").json()

    def financialStatement(
        self, 
        statement:str="summary",
        period:str="annual",
        reportType:str="original",
        export:bool=False,
        folderPath:str="."
    ) -> dict:
        """
        This function retrieves the financial statement or downloads it as a xls file.

        Args:
            statement (str) : possible values are balancesheet, cashflow, incomestatement, summary
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated
            export (bool) : if True, the file is downloaded in the folderPath
            folderPath (str) : path to save the file if export is True

        Returns:
            dict with financial statement

        Examples:
            >>> Stock("US0378331005").financialStatement('summary', 'quarterly', 'original')
            >>> Stock("US0378331005").financialStatement('cashflow', 'annual', 'restated')
            >>> Stock("US0378331005").financialStatement('balancesheet', 'annual', 'restated')
            >>> Stock("US0378331005").financialStatement('incomestatement', 'annual', 'restated')

        """

        if not isinstance(statement, str):
            raise TypeError("statement parameter should be a string")

        if not isinstance(period, str):
            raise TypeError("period parameter should be a string")

        if not isinstance(reportType, str):
            raise TypeError("reportType parameter should be a string")
        
        if not isinstance(export, bool):
            raise TypeError("export parameter should be a boolean")
        
        if not isinstance(folderPath, str):
            raise TypeError("folderPath parameter should be a string")

        statement_choice = {
            "balancesheet": "balanceSheet",
            "cashflow": "cashFlow",
            "incomestatement": "incomeStatement",
            "summary": "summary",
        }

        period_choice = {"annual": "A", "quarterly": "Q"}

        reportType_choice = {"original": "A", "restated": "R"}

        if statement not in statement_choice:
            raise ValueError(
                f"statement parameter must take one of the following value : { ', '.join(statement_choice)}"
            )

        if period not in period_choice:
            raise ValueError(
                f"period parameter must take one of the following value : { ', '.join(period_choice)}"
            )

        if reportType not in reportType_choice:
            raise ValueError(
                f"reportType parameter must take one of the following value : { ', '.join(reportType_choice.keys())}"
            )
        


        params = {"reportType": reportType_choice[reportType]}

        if export:
            params["operation"] = "export"

        if statement == "summary":
            response = self.GetData(
                "newfinancials", params=params, url_suffix=f"{period}/summary"
            )

        params["dataType"] = period_choice[period]

        response = self.GetData(
            "newfinancials",
            params=params,
            url_suffix=f"{statement_choice[statement]}/detail",
        )

        if not export:
            return response.json()
        
        timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        fileName = f"{statement_choice[statement]}-{self.code}-{timestamp}.xls"
        with open(f"{folderPath}/{fileName}", "wb") as f:
            f.write(response.content)

        return {"status" : response.status_code, 
                "mesage" : "File downloaded", 
                "filename" : fileName,
                "folder" : folderPath}

    def financialSummary(self, 
                         period:str="annual", 
                         reportType:str="original",
                        ) -> dict:
        """
        This function retrieves the financial statement summary.

        Args:
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated
            export (bool) : if True, the file is downloaded in the folderPath
            folderPath (str) : path to save the file if export is True

        Returns:
            dict with financial statement summary

        Examples:
            >>> Stock("US0378331005").financialSummary('quarterly', 'original')
            >>> Stock("US0378331005").financialSummary('annual', 'restated')

        """

        return self.financialStatement("summary", 
                                       period=period, 
                                       reportType=reportType)

    def freeCashFlow(self) -> dict:
        """
        This function retrieves the free cash flow.

        Returns:
            dict with free cash flow

        Examples:
            >>> Stock("US0378331005").freeCashFlow()

        """
        return self.GetData("keyMetrics/cashFlow", url_suffix="").json()

    def historical(self,
                   start_date:datetime.datetime, 
                   end_date:datetime.datetime, 
                   frequency:str="daily") -> list:
        """
        This function retrieves the historical price, volume and dividends of the stock.

        Args:
            start_date (datetime) : start date to get history
            end_date (datetime) : end date to get history
            frequency (str) : can be daily, weekly, monthly

        Returns:
            list of dict with price, volume and dividend

        Examples:
            >>> Stock("US0378331005").history(datetime.datetime.today()- datetime.timedelta(30),datetime.datetime.today())

        """
        return self.TimeSeries(
            ["open", "high", "low", "close", "volume", "previousClose", "dividend"],
            start_date=start_date,
            end_date=end_date,
            frequency=frequency,
        )

    def incomeStatement(self, 
                        period:str="annual", 
                        reportType:str="original",
                        export:bool=False,
                        folderPath:str="."
                        ) -> dict:
        """
        This function retrieves the income statement.

        Args:
            period (str) : possible values are annual, quarterly
            reportType (str) : possible values are original, restated
            export (bool) : if True, the file is downloaded in the folderPath
            folderPath (str) : path to save the file if export is True

        Returns:
            dict with income statement

        Examples:
            >>> Stock("US0378331005").incomeStatement('quarterly', 'original')
            >>> Stock("US0378331005").incomeStatement('annual', 'restated')

        """

        return self.financialStatement(
            "incomestatement", 
            period=period, 
            reportType=reportType,
            export=export,
            folderPath=folderPath
            )

    def institutionBuyers(self,
                          top:int=20) -> dict:
        """
        This function retrieves the institutions which buy the stock.

        Args:
            top (int) : number of institutions to return
        Returns:
            dict with the buyers

        Examples:
            >>> Stock("US0378331005").institutionBuyers(top=50)

        """

        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData(
            "ownership/v1", url_suffix=f"Buyers/institution/{top}/data"
        ).json()

    def institutionConcentratedOwners(self, 
                                      top:int=20) -> dict:
        """
        This function retrieves the institutions which are concentrated on the stock.

        Args:
            top (int) : number of institutions to return
        Returns:
            dict with the concentarted owners

        Examples:
            >>> Stock("US0378331005").institutionConcentratedOwners(top=50)

        """

        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData(
            "ownership/v1", url_suffix=f"ConcentratedOwners/institution/{top}/data"
        ).json()

    def institutionOwnership(self, 
                             top:int=20) -> dict:
        """
        This function retrieves the main institutions which own the stock.

        Args:
            top (int) : number of institutions to return
        Returns:
            dict with the main owners

        Examples:
            >>> Stock("US0378331005").institutionOwnership(top=50)

        """

        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData(
            "ownership/v1", url_suffix=f"OwnershipData/institution/{top}/data"
        ).json()

    def institutionSellers(self, 
                           top:str=20) -> dict:
        """
        This function retrieves the institutions which sell on the stock.

        Args:
            top (int) : number of institutions to return
        Returns:
            dict with sellers

        Examples:
            >>> Stock("US0378331005").institutionSellers(top=50)

        """
        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData(
            "ownership/v1", url_suffix=f"Sellers/institution/{top}/data"
        ).json()

    def keyExecutives(self) -> dict:
        """
        This function retrieves information oabout key excutives of the company.

        Returns:
            dict with key executives information

        Examples:
            >>> Stock("US0378331005").keyExecutives()

        """
        return self.GetData("insiders/keyExecutives").json()

    def keyMetricsSummary(self, 
                          reportType:str="original") -> dict:
        """
        This function retrieves the key metrics summary

        Args:
            reportType (str) : possible values are original, restated

        Returns:
            dict with key metrics summary

        Examples:
            >>> Stock("US0378331005").keyMetricsSummary()

        """
        if not isinstance(reportType, str):
            raise TypeError("reportType parameter should be a string")
        
        reportType_choice = {"original": "A", "restated": "R"}

        if reportType not in reportType_choice:
            raise ValueError(
                f"reportType parameter must take one of the following value : { ', '.join(reportType_choice.keys())}"
            )

        params = {"reportType": reportType_choice[reportType]}

        return self.GetData("keyMetrics/summary",
                             params=params,
                            url_suffix="").json()

    def keyRatio(self) -> dict:
        """
        This function retrieves the key ratio of the stock.

        Returns:
            dict with key ratio

        Examples:
            >>> Stock("US0378331005").keyRatio()

        """
        return self.GetData("keyratios").json()

    def mutualFundBuyers(self,
                         top:int=20) -> dict:
        """
        This function retrieves the mutual funds which buy the stock.

        Args:
            top (int) : number of mutual funds to return
        Returns:
            dict with the buyers

        Examples:
            >>> Stock("US0378331005").mutualFundBuyers(top=50)

        """

        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData("ownership/v1", url_suffix=f"Buyers/mutualfund/{top}/data").json()

    def mutualFundConcentratedOwners(self, 
                                     top:int=20) -> dict:
        """
        This function retrieves the mutual funds which are concentrated on the stock.

        Args:
            top (int) : number of mutual funds to return
        Returns:
            dict with the concentarted owners

        Examples:
            >>> Stock("US0378331005").mutualFundConcentratedOwners(top=50)

        """

        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData(
            "ownership/v1", url_suffix=f"ConcentratedOwners/mutualfund/{top}/data"
        ).json()

    def mutualFundOwnership(self, 
                            top:int=20) -> dict:
        """
        This function retrieves the main mutual funds which own the stock.

        Args:
            top (int) : number of mutual funds to return
        Returns:
            dict with the main owners

        Examples:
            >>> Stock("US0378331005").mutualFundOwnership(top=50)

        """

        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData(
            "ownership/v1", url_suffix=f"OwnershipData/mutualfund/{top}/data"
        ).json()

    def mutualFundSellers(self,
                          top:int=20) -> dict:
        """
        This function retrieves the mutual funds which sell on the stock.

        Args:
            top (int) : number of mutual funds to return
        Returns:
            dict with sellers

        Examples:
            >>> Stock("US0378331005").mutualFundSellers(top=50)

        """

        if not isinstance(top, int):
            raise TypeError("top parameter should be an integer")

        return self.GetData(
            "ownership/v1", url_suffix=f"Sellers/mutualfund/{top}/data"
        ).json()

    def operatingGrowth(self) -> dict:
        """
        This function retrieves the operating growth of the stock.

        Returns:
            dict with operating growth

        Examples:
            >>> Stock("US0378331005").operatingGrowth()

        """
        return self.GetData("keyStats/growthTable", url_suffix="").json()
    
    def overview(self) -> dict:
        """
        This function retrieves stock overview.

        Returns:
            dict with stock overview

        Examples:
            >>> Stock("US0378331005").overview()

        """
        return self.GetData("equityOverview").json()


    def profitability(self) -> dict:
        """
        This function retrieves the profitability of the stock.

        Returns:
            dict with profitability ratios

        Examples:
            >>> Stock("US0378331005").profitability()

        """
        return self.GetData("keyMetrics/profitabilityAndEfficiency", url_suffix="").json()

    def sustainability(self) -> dict:
        """
        This function retrieves the sustainability of the stock.

        Returns:
            dict with sustainability

        Examples:
            >>> Stock("US0378331005").sustainability()

        """
        return self.GetData("esgRisk/sustainability").json()
    
    def split(self) -> dict:
        """
        This function retrieves the split history of the stock.

        Returns:
            dict with split history

        Examples:
            >>> Stock("US0378331005").split()

        """
        return self.GetData("split/v1").json()

    def tradingInformation(self) -> dict:
        """
        This function retrieves the trading information of the stock
        such as Previous Close Price, Day Range, 52-Week Range, Bid/Ask,
        Market Cap, Volume/Avg
        Returns:
            dict with performance
        Examples:
            >>> Stock("US0378331005").tradingInformation()

        """
        return self.RealtimeData("quotes")
    
    def trailingTotalReturn(self) -> dict:
        """
        This function retrieves the performance of the stock and its index.

        Returns:
            dict with performance

        Examples:
            >>> Stock("US0378331005").trailingTotalReturn()

        """
        return self.GetData("trailingTotalReturns").json()

    def transactionHistory(self) -> list:
        """
        This function retrieves the transaction of key people.

        Returns:
            list o dict of transaction of key people

        Examples:
            >>> Stock("US0378331005").transactionHistory()

        """
        return self.GetData("insiders/transactionHistory").json()

    def transactionSummary(self) -> list:
        """
        This function retrieves the summuary of transactions of key people.

        Returns:
            list of dict with transactions

        Examples:
            >>> Stock("US0378331005").transactionSummary()

        """
        return self.GetData("insiders/transactionChart").json()

    def valuation(self) -> dict:
        """
        This function retrieves the valution of the stock.

        Returns:
            dict with valuation

        Examples:
            >>> Stock("US0378331005").valuation()

        """
        return self.GetData("valuation/v3", url_suffix="").json()
    



import requests
import re
import warnings
import pandas as pd

from .utils import random_user_agent
from .utils import ASSET_TYPE, FILTER_TYPE, LANGUAGE
from .error import not_200_response

def general_search(params:dict,
                   language:str="en-gb", 
                   proxies:dict=None) -> dict:
    """
    This function will use the screener of morningstar.com
    to find informations about funds or classification

    Args:
      params (dict) : paramaters of the request
      language (str) : language of the request, default is "en-gb"
      proxies (dict) : set the proxy if needed,
      example : {"http": "http://host:port","https": "https://host:port"}

    Returns:
      dict of information

    Examples:
      >>> general_search(params = {
                                    'query': "_ ~= 'US67066G1040'", 
                                    'fields': 'isin,name', 
                                    'limit': 3
                                    })
    """

    if not isinstance(params, dict):
        raise TypeError("params parameter should be dict")
    
    if not isinstance(language, str):
        raise TypeError("language parameter should be a string")

    if proxies and not isinstance(proxies, dict):
        raise TypeError("proxies parameter should be dict")
    
    if language not in LANGUAGE:
        raise ValueError(
            f"language parameter can only take one of the values : {', '.join(LANGUAGE)}"
        )
    # url
    url = f"https://global.morningstar.com/api/v1/{language}/tools/screener/_data"
    # headers
    headers = {
        "user-agent": random_user_agent(),
    }

    response = requests.get(url, params=params, headers=headers, proxies=proxies)

    not_200_response(url, response)

    return response.json()

def screener_universe(
    term:str,
    language:str="en-gb",
    field:str|list="",
    filters:dict=None,
    pageSize:int=10,
    page:int=1,
    sortby:str=None,
    ascending:bool=True,
    proxies:dict=None
    ) -> list:
    """
    This function will use the screener of global.morningstar.com
    to find funds, etf, stocks which include the term.

    Args:
      term (str): text to find a security can be a the name, 
      part of a name or the isin
      language (str): language of the request, default is "en-gb"
      field (str | list) : field to find
      filters (dict) : filter, use the method search_filter() to find the different possible filter keys
      pageSize (int): number of securities to return
      page (int): page to return
      sortby (str) : sort by a field
      ascending (bool) : True sort by ascending order, False sort by descending order
      proxies (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}

    Returns:
      list of dict with secrity information
        [{'meta': {'securityID': 'F00000MRIF', 'performanceID': '0P0000TUB0', 'fundID': 
        'FS00008MVC', 'masterPortfolioID': '2852260', 'universe': 'FO'}, 
        'fields': {'isin': {'value': 'FR0010921445'}, 
        'name': {'value': 'Abeille Capital Planète'}}}, 
        {'meta': {'securityID': 'FOUSA06JVV', 'performanceID': 
        '0P00009W2T', 'fundID': 'FSUSA08EHM', 'masterPortfolioID': '237848', 'universe': 'FO'}, 
        'fields': {'isin': {'value': 'FR0010234898'}, 'name': 
        {'value': 'Candriam MM Long/Short Global C EUR'}}}, 
        {'meta': {'securityID': 'FOUSA06JU1', 'performanceID': '0P00009W0Z', 
        'fundID': 'FSUSA08EHM', 'masterPortfolioID': '237848', 'universe': 'FO'},
        'fields': {'isin': {'value': 'FR0000974412'}, 'name': 
        {'value': 'Candriam MM Long/Short Global F EUR'}}}]

    Examples:
      >>> screener_universe("myria",field=["isin", "name"],pageSize=10,page=1, sortby="name", ascending=False)
      >>> screener_universe("US67066G1040", language="de")

    """
    if not isinstance(term, str):
        raise TypeError("term parameter should be a string")
    
    if not isinstance(language, str):
        raise TypeError("language parameter should be a string")
    
    if not isinstance(field, (str, list)):
        raise TypeError("field parameter should be a string or a list")
    
    if filters and not isinstance(filters, dict):
        raise TypeError("filters parameter should be a dict")
    
    if not isinstance(pageSize, int):
        raise TypeError("pageSize parameter should be an integer")
    
    if not isinstance(page, int):
        raise TypeError("page parameter should be an integer")
        
    if sortby and not isinstance(sortby, str):
        raise TypeError("sortby parameter should be a string")
    
    if not isinstance(ascending, bool):
        raise TypeError("ascending parameter should be a boolean")
    

    if proxies and not isinstance(proxies, dict):
        raise TypeError("proxies parameter should be dict")
    
    if language not in LANGUAGE:
        raise ValueError(
            f"language parameter can only take one of the values : {', '.join(LANGUAGE)}"
        )

    all_fields = search_field(display_print=False)
    if not field:
        check_field = True
        fields = field
    elif isinstance(field, str):
        check_field = field in all_fields
        fields = field
    else:
        check_field = set(field).issubset(set(all_fields))
        fields = ",".join(field)

    if sortby and sortby not in all_fields:
        raise ValueError(
            f"""The sort field {sortby} is not a valid field.
            You can find the possible fields with the method search_field().
            Possible fields are : {', '.join(all_fields)}"""
        )

    if not check_field:
        raise ValueError(
            f"""The field {field} is not a valid field.
            You can find the possible fields with the method search_field().
            Possible fields are : {', '.join(all_fields)}"""
        )
    
    query_params = f"_ ~= '{term}'"

    if filters:
        list_filter =search_filter()
        for f in filters:
            if f not in list_filter:
                warnings.warn(
                    f"""{f} is not a valid filter and will be ignored.
                    You can find the possible filters with the method search_filter()."""
                )
            else:
                # if list, IN condition
                if isinstance(filters[f], list):
                    query_params += f""" AND {f} IN ({','.join(f"'{x}'" for x in filters[f])})"""
                # if tuple, either, < or > condition
                elif isinstance(filters[f], tuple):
                    if len(filters[f]) != 2:
                        warnings.warn(f"""{f} is not a valid filter and will be ignored.
                                      The tuple has to be of a length of 2""")
                    if filters[f][0] not in ['<','<=','>=', '>']:
                        warnings.warn(f"""{f} is not a valid filter and will be ignored.
                                       The first argument of the tuple has to be one of this value {','.join(['<','<=','>=', '>'])}""")
                        continue
                    if not isinstance(filters[f][1],(int,float)):
                        warnings.warn(f"""{f} is not a valid filter and will be ignored.    
                                        The second argument of the tuple has be a number.
                                            """)
                        continue
                    query_params += f" AND {f} {filters[f][0]} {filters[f][1]}"
                # else = condition
                else:
                    query_params += f" AND {f} = '{filters[f]}'"
    params = {
        "query": query_params,
        "fields" : fields,
        "page" : page,
        "limit": pageSize,
    }   
    if sortby:
        if ascending == True:
            params["sort"] = f"{sortby}:asc"
        else:
            params["sort"] = f"{sortby}:desc"


    

    result = general_search(params, 
                            language=language,
                            proxies=proxies)

    if not "results" in result:
        print(f"0 fund found whith the term {term}")
        return {}
    
    return result["results"]

def search_field(pattern:str="",
                 display_print=True) -> list:
    """
    This function retrieves the possible fields for the screener 

    Args:
    pattern (str) : text contained in the field
    display_print (bool) : if True, print the possible fields

    Returns:
        list of possible fields for the screener of securities

    Example:
        >>> search_field('feE')
        >>> search_field('reTurn')

    """
    if not isinstance(pattern, str):
        raise TypeError("pattern parameter should be a string")
    
    headers = {"user-agent": random_user_agent()}
    
    url = "https://global.morningstar.com/api/v1/fr/stores/data-points/fields"

    response = requests.get(url, headers=headers)
    not_200_response(url, response)
    if "results" not in response.json():
        raise ValueError("No results found for the given pattern")
    
    result = response.json()["results"]
    df = pd.DataFrame(result)
    df_filtered = df.loc[df["field"].str.contains(f"(?i){pattern}",regex=True)]
    filtered_list = df_filtered["field"].tolist()

    if display_print:
        print(f"possible fields for screener can be : {', '.join(filtered_list)}")

    return filtered_list



def search_filter(pattern:str="",
                asset_type:str="",
                filter_type:str="",
                explicit:bool=False) -> list:
                  
    """
    This function retrieves the possible filters for the parameter filters of the function screener_universe

    Args:
        pattern (str): text contained in the field
        asset_type (str): type of asset, can be one of the values in ASSET_TYPE
        filter_type (str): type of filter, can be one of the values in FILTER_TYPE
        explicit (bool): if True, return a list of dict with fields and their metadata
    Returns:
        list of possible filters
        Raise[{'field': 'dividendYield', 'label': 'Dividend Yield (%)', 
        'numeric': True, 'type': 'range', 'options': 
        {'type': [{'text': 'Forward', 'value': 'forward'}, 
        {'text': 'Trailing', 'value': 'trailing', 'default': True}]}}]

        ['stockStyleBox', 'fundEquityStyleBox', 'fundFixedIncomeStyleBox', 'fundAlternativeStyleBox']
    Example:
        >>> search_filter()
        >>> search_filter(pattern="div",asset_type="stock", filter_type="dividends", explicit=True)

    """

    if not isinstance(pattern, (str)):
        raise TypeError("pattern parameter should be a string")
    if not isinstance(asset_type, (str)):
        raise TypeError("asset_type parameter should be a string")
    
    if asset_type and asset_type not in ASSET_TYPE:
        raise TypeError(
            f"asset_type parameter can only take one of the values : {','.join(ASSET_TYPE)}"
        )
    if not isinstance(filter_type, (str)):
        raise TypeError("filter_type parameter should be a string")
    
    if filter_type and filter_type not in FILTER_TYPE:
        raise TypeError(
            f"filter_type parameter can only take one of the values : {','.join(FILTER_TYPE)}"
        )
    
    headers = {"user-agent": random_user_agent()}
    url = "https://global.morningstar.com/api/v1/fr/stores/filters"

    response = requests.get(url, headers=headers)
    not_200_response(url, response)

    result = response.json()

    list_filter = ["investmentType","countriesOfSale"]
    list_filter_explicit = []
    for security_type in result["results"]:
        if asset_type:
            if security_type["id"] != f"{asset_type}-filters":
                continue
        security_filter_type = security_type["filters"]
        for filters in security_filter_type:
            if filter_type:
                if filter_type != filters["id"]:
                    continue
            for child in filters["children"]:
                if pattern:
                    if not re.search(pattern, child["field"], re.IGNORECASE):
                        continue
                if child["field"] in list_filter:
                    continue
                list_filter.append(child["field"])
                list_filter_explicit.append(child)
    if explicit:
        return list_filter_explicit
    return list_filter

def token_chart(proxies:dict=None) -> str:
    """
    This function will scrape the Bearer Token needed to access MS API chart data

    Args:
    proxies (dict) : set the proxy if needed ,
    example : {"http": "http://host:port","https": "https://host:port"}

    Returns:
    str bearer token

    """

    if proxies and not isinstance(proxies, dict):
        raise TypeError("proxies parameter should be dict")

    url = "https://www.morningstar.com/funds/xnas/afozx/chart"

    headers = {"user-agent": random_user_agent()}

    response = requests.get(url, headers=headers, proxies=proxies)

    all_text = response.text
    if all_text.find("token") == -1:
        return None

    token_start = all_text[all_text.find("token") :]
    return token_start[7 : token_start.find("}") - 1]




import datetime
import re
import requests

from .error import not_200_response

from .search import (screener_universe,
                     token_chart
                     )
from .utils import (
    APIKEY,
    random_user_agent,
    LANGUAGE
    )



class Security:
    """
    Parent class to access data about security

    Args:
        term (str): text to find a fund can be a name, part of a name or the isin of the funds
        asset_type (str) : security type from the inherited, can be fund, stock, etf
        filters (dict) : filter, use the method search_filter() to find the different possible filter keys
        pageSize (int): number of securities to return
        page (int): page to return
        sortby (str) : sort by a field
        ascending (bool) : True sort by ascending order, False sort by descending order
        itemRange (int) : index of funds to return (must be inferior to PageSize)
        proxies (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}

    Examples:
        >>> Security('0P0000712R', "fund", 9, 0)
        >>> Security('visa', "stock", 25, 2)

    Raises:
        TypeError: raised whenever the parameter type is not the type expected
        ValueError : raised whenever the parameter is not valid or no fund found

    """

    def __init__(
        self,
        term:str,
        language:str="en-gb",
        asset_type:str="",
        filters:dict=None,
        itemRange:int=0,
        pageSize:int=10,
        page:int=1,
        sortby:str=None,
        ascending:bool=True,
        proxies:dict=None,
    ) -> None:
        
        if not isinstance(term, str):
            raise TypeError("term parameter should be a string")
        
        if not isinstance(language, str):
            raise TypeError("language parameter should be a string")
        
        if not isinstance(asset_type, str):
            raise TypeError("asset_type parameter should be a string")

        if not isinstance(pageSize, int):
            raise TypeError("pageSize parameter should be an integer")
        
        if not isinstance(page, int):
            raise TypeError("page parameter should be an integer")
        
        if sortby and not isinstance(sortby, str):
            raise TypeError("sortby parameter should be a string")
    
        if not isinstance(ascending, bool):
            raise TypeError("ascending parameter should be a boolean")

        if not isinstance(itemRange, int):
            raise TypeError("itemRange parameter should be an integer")

        if pageSize <= itemRange:
            raise ValueError(
                "itemRange parameter should be strictly inferior to pageSize parameter"
            )

        if filters and not isinstance(filters, dict):
            raise TypeError("filters parameter should be dict")

        if proxies and not isinstance(proxies, dict):
            raise TypeError("proxies parameter should be dict")
        
        if language not in LANGUAGE:
            raise ValueError(
                f"language parameter can only take one of the values : {', '.join(LANGUAGE)}"
            )


        self.language = language
        self.proxies = proxies
        self.filters = filters
        self.pageSize = pageSize
        self.page = page
        self.sortby = sortby
        self.ascending = ascending
        self.itemRange = itemRange

        self.asset_type = "security"

        code_list = []

        code_list = screener_universe(
                term,
                language=self.language,
                field=["isin", "name"],
                filters=filters,
                pageSize=pageSize, 
                page=page,
                sortby=sortby,
                ascending=ascending,
                proxies=self.proxies,)
        

        if code_list:
            if itemRange < len(code_list):
                self.code = code_list[itemRange]['meta']["securityID"]
                self.name = code_list[itemRange]['fields']["name"]['value']
                self.isin = code_list[itemRange]['fields']["isin"]['value']
                universe = code_list[itemRange]['meta']["universe"]

                if universe == "EQ":
                    self.asset_type = "stock"
                elif universe == "FE":
                    self.asset_type = "etf"
                elif universe == "FO":
                    self.asset_type = "fund"

                if universe == "EQ" and asset_type in ["etf", "fund"]:
                    raise ValueError(
                        f"The security found with the term {term} is a stock and the parameter asset_type is equal to {asset_type}, the class Stock should be used with this security."
                    )

                if universe in ["FO", "FE"] and asset_type == "stock":
                    if universe == "FO":
                        raise ValueError(
                            f"The security found with the term {term} is a fund and the parameter asset_type is equal to {asset_type}, the class Fund should be used with this security."
                        )
                    else:
                        raise ValueError(
                            f"The security found with the term {term} is an ETF and the parameter asset_type is equal to {asset_type}, the class Fund should be used with this security."
                        )

            else:
                raise ValueError(
                    f"Found only {len(code_list)} {self.asset_type} with the term {term}. The paramater itemRange must maximum equal to {len(code_list)-1}"
                )
        else:
            raise ValueError(f"0 {self.asset_type} found with the term {term}")



    def dataPoint(self, 
                  field:str|list) -> list[dict]:
        """
        This function retrieves infos about securities such as name,
        performance, risk metrics...

        Args:
        field (str or list) : field to find

        Returns:
            list of dict security infos

        Example:
            >>> Security("myria").dataPoint(['name', 'isin', 'priipsKidCosts'])
            >>> Security("myria").dataPoint('standardDeviation')

        """

        result =  screener_universe(
            self.isin, 
            language=self.language,
            field=field,
            filters=self.filters,
            proxies=self.proxies,
            pageSize=self.pageSize,
            page=self.page,
            sortby=self.sortby,
            ascending=self.ascending,
        )
        
        return result[self.itemRange]['fields']
        
    
    def GetData(self, 
                field:str, 
                params:dict=None, 
                headers:dict=None, 
                url_suffix:str="data") -> dict|list:
        """
        This function retrieves data from the MorningStar global API.
        Args:
            field (str) : endpoint of the request
            params (dict) : parameter for the request
            headers (dict) : headers of the request
            url_suffix (str) : suffix of the url

        Raises:
            TypeError raised whenever type of paramater are invalid

        Returns:
            dict with data

        Examples:
            >>> Security("rmagx").GetData("price/feeLevel")

        """

        if not isinstance(field, str):
            raise TypeError("field parameter should be a string")

        if params and not isinstance(params, dict):
            raise TypeError("params parameter should be a dict")

        if not isinstance(url_suffix, str):
            raise TypeError("url_suffix parameter should be a string")
        
        if headers and not isinstance(headers, dict):
            raise TypeError("headers parameter should be a dict")

        # url of API
        url = f"""https://api-global.morningstar.com/sal-service/v1/{self.asset_type}/{field}/{self.code}"""

        if url_suffix:
            url += f"""/{url_suffix}"""

        # headers
        default_headers = {
            "apikey": APIKEY,
        }
        if headers:
            default_headers = default_headers | headers

        #params 
        default_params = {
            "clientId": "MDC",
            "version": "4.71.0"

        }
        if params:
            default_params = default_params | params


        response = requests.get(
            url, params=default_params, headers=default_headers, proxies=self.proxies
        )

        not_200_response(url, response)

        return response

    def ltData(self, 
               field:str, 
               currency:str="EUR") -> dict:
        """
        Generic function to use MorningStar lt api.

        Args:
            field (str) : viewId in the params
            currency (str) : currency in 3 letters

        Raises:
            TypeError raised whenever type of paramater are invalid

        Returns:
            dict with data

        Examples:
            >>> Security("rmagx").ltData("price/feeLevel")

        """
        if not isinstance(field, str):
            raise TypeError("field parameter should be a string")

        # url of API
        url = f"""https://lt.morningstar.com/api/rest.svc/klr5zyak8x/security_details/{self.code}"""

        params = {
            "viewId": field,
            "currencyId": currency,
            "itype": "msid",
            "languageId": "en",
            "responseViewFormat": "json",
        }
        response = requests.get(url, params=params, proxies=self.proxies)

        not_200_response(url, response)

        # responseis a list
        response_list = response.json()
        if response_list:
            return response_list[0]
        else:
            return {}

    def RealtimeData(self, 
                     url_suffix: str) -> dict:
        """
        This function retrieves historical data of the specified fields

        Args:
            url_suffix (str) : suffixe of the url

        Returns:
            dict of realtime data

        Examples:
            >>> Security("visa").RealtimeData("quotes")

        Raises:
            TypeError: raised whenever the parameter type 
            is not the type expected
            ConnectionError : raised whenever the response is not 200 OK

        """
        # error raised if url_suffix is not a string
        if not isinstance(url_suffix, str):
            raise TypeError("url_suffix parameter should be a string or a list")
        # url for realtime data
        url = f"""https://www.morningstar.com/api/v2/stores/realtime/{url_suffix}"""

        # header with user agent
        headers = {
                    'user-agent': random_user_agent(), 
                    }
        #parameters of the request
        params = {"securities": self.code}
        # response
        response = requests.get(url, 
                                params=params, 
                                headers=headers, 
                                proxies=self.proxies,
                                timeout=60)
        # manage response
        not_200_response(url, response)
        # result
        return response.json()
    
    def TimeSeries(self, 
                   field:str|list, 
                   start_date:datetime.datetime,
                   end_date:datetime.datetime,
                   frequency:str="daily") -> list:
        """
        This function retrieves historical data of the specified fields

        Args:
            field (str|list) : field to retrieve, can be a string or a list of string
            start_date (datetime) : start date to get history
            end_date (datetime) : end date to get history
            frequency (str) : can be daily, weekly, monthly

        Returns:
            list of dict time series

        Examples:
            >>> Security("RMAGX").TimeSeries(["nav","totalReturn"],datetime.datetime.today()- datetime.timedelta(30),datetime.datetime.today())

        Raises:
            TypeError: raised whenever the parameter type is not the type expected
            ValueError : raised whenever the parameter is not valid or no funds found

        """

        # error raised if field is not a string or a list
        if not isinstance(field, (str, list)):
            raise TypeError("field parameter should be a string or a list")

        # error raised if start_date is note a datetime.date
        if not isinstance(start_date, datetime.date):
            raise TypeError("start_date parameter should be a datetime.date")

        # error raised if end_date is note a datetime.date
        if not isinstance(end_date, datetime.date):
            raise TypeError("end_date parameter should be a datetime.date")

        # error if end_date < start_date
        if end_date < start_date:
            raise ValueError("end_date must be more recent than start_date")

        # error raised if frequency is not a string
        if not isinstance(frequency, str):
            raise TypeError("frequency parameter should be a string")

        # dict of frequency
        frequency_row = {"daily": "d", "weekly": "w", "monthly": "m"}

        if self.asset_type == "stock":
            frequency_row = {"daily": "d", 
                             "5min": "5", 
                             "10min": "10", 
                             "15min": "15", 
                             "30min": "30",
                             }
        else:
            frequency_row = {"daily": "d", "monthly": "m", "quarterly": "q"}


        # raise an error if frequency is not daily, wekly or monthly
        if frequency not in frequency_row:
            raise ValueError(
                f"frequency parameter must take one of the following value : { ', '.join(frequency_row.keys())}"
            )

        if isinstance(field, list):
            queryField = ",".join(field)
        else:
            queryField = field

        # bearer token
        bearer_token = token_chart()
        # url for nav
        url = "https://www.us-api.morningstar.com/QS-markets/chartservice/v2/timeseries"
        # header with bearer token
        headers = {
            "user-agent": random_user_agent(),
            "authorization": f"Bearer {bearer_token}",
        }
        #params of the request
        params ={
            "query" : f"{self.code}:{queryField}",
            "frequency": frequency_row[frequency],
            "startDate": start_date.strftime('%Y-%m-%d'),
            "endDate": end_date.strftime('%Y-%m-%d'),
            "trackMarketData": "3.6.3",
            "instid": "DOTCOM",
        }
        # response
        response = requests.get(url,
                                params=params,
                                headers=headers, 
                                proxies=self.proxies)
        # manage response
        not_200_response(url, response)
        # result
        result = response.json()
        # return empty list if we don't get data
        if not result:
            return []
        if "series" in result[0]:
            return result[0]["series"]

        return []



from setuptools import setup, find_packages
import io

def readme():
    with io.open('README.md', encoding='utf-8') as f:
        return f.read()

def requirements(filename):
    reqs = list()
    with io.open(filename, encoding='utf-8') as f:
        for line in f.readlines():
            reqs.append(line.strip())
    return reqs



setup(
    name="mstarpy",
    packages=find_packages(),
    version="7.0.1",
    python_requires='>=3.10',
    license='MIT License',
    author="Maël Jourdain",
    author_email='mael.jourdain@gmail.com',
    description="Mutual funds and stocks data extraction from MorningStar with Python",
    long_description= readme(),
    long_description_content_type="text/markdown",
    install_requires=requirements(filename='requirements/requirements.txt'),
    include_package_data=True,
)


"""module to raise error"""
import requests

def not_200_response(url:str, 
                     response:requests.models.Response) -> None:
    """
    This function raise a ConnectionError
    if the status code a requests is not 200.
    """
    if not response.status_code == 200:
        raise ConnectionError(
            f"""Error {response.status_code}
            for the api {url}. Message : {response.reason}."""
        )


def no_site_error(code:str,
                  name: str,
                  country: str, 
                  site:str) -> None:
    """
    This function raise a ValueError if the
    selected country is "us" or a site is not selected.
    """
    if not site or country == "us":
        if country:
            raise ValueError(f"""The funds of the
                             country {country} cannot be scraped.""")
        raise ValueError(f"The funds {name} ({code}) cannot be scraped.")



""" class funds """
import pandas as pd
import datetime
import requests
import warnings

from .search import screener_universe
from .security import Security
from .utils import random_user_agent


class Funds(Security):
    """
    Main class to access data about funds and etf, inherit from Security class

    Args:
        term (str): text to find a fund, can be a name, part of a name or the isin of the funds
        language (str): language of the data, default is "en-gb"
        filters (dict) : filter, use the method search_filter() to find the different possible filter keys
        itemRange (int) : index of stocks to return (must be inferior to PageSize)
        pageSize (int): number of securities to return
        page (int): page to return
        sortby (str) : sort by a field
        ascending (bool) : True sort by ascending order, False sort by descending order
        proxies (dict) : set the proxy if needed , example : {"http": "http://host:port","https": "https://host:port"}

    Examples:
        >>> Funds("myria",page=2, pageSize=15,itemRange=3,sortby="name", ascending=True)
        >>> Funds('bond', 25, 2)

    Raises:
        TypeError: raised whenever the parameter type is not the type expected
        ValueError : raised whenever the parameter is not valid or no fund found

    """

    def __init__(
        self,
        term:str,
        language:str="en-gb",
        filters:dict=None,
        itemRange:int=0,
        pageSize:int=10,
        page:int=1,
        sortby:str=None,
        ascending:bool=True,
        proxies:dict=None,
    ) -> None:
        
        fund_filter = {"investmentType" : ['FE', 'FO']}
        if filters:
            fund_filter = fund_filter | filters
        
        super().__init__(
            term=term,
            asset_type="fund",
            language=language,
            filters=fund_filter,
            itemRange=itemRange,
            pageSize=pageSize,
            page=page,
            sortby=sortby,
            ascending=ascending,
            proxies=proxies,
        )

    def allocationMap(self, 
                       version:int=3) -> dict:
        """
        This function retrieves the asset allocation of the funds, index and category.
        Args:
            version (int) : version of the api of allocation map 2 or 3
        Returns:
            dict with allocation map

        Examples:
            >>> Funds("myria").allocationMap()

        """
        if not isinstance(version,int):
            raise TypeError("version paramater should be an integer")
        
        if version not in range(2,4):
            raise ValueError("version paramater should be 2 or 3")

        return self.GetData(f"process/asset/v{version}").json()
    

    def allocationWeighting(self) -> dict:
        """
        This function retrieves the Growth/Blend/Value and
        market capitalizaton allocation size of the funds.

        Returns:
            dict with allocation

        Examples:
            >>> Funds("myria").allocationWeighting()

        """
        return self.GetData("process/weighting").json()

    def analystRating(self) -> list[dict]:
        """
        This function retrieves the rating of the funds

        Returns:
            list of dict with ratings

            >>> Funds("myria").analystRating()

        """

        return self.GetData("parent/analystRating").json()

    def analystRatingTopFunds(self) -> dict:
        """
        This function retrieves the rating Top funds

        Returns:
            dict with ratings

            >>> Funds("myria").analystRatingTopFunds()

        """

        return self.GetData("parent/analystRating/topfunds").json()

    def analystRatingTopFundsUpDown(self) -> dict:
        """
        This function retrieves the rating funds Up Down

        Returns:
            dict with ratings

            >>> Funds("myria").analystRatingTopFundsUpDown()

        """

        return self.GetData("parent/analystRating/topfundsUpDown").json()


    def carbonMetrics(self) -> dict:
        """
        This function retrieves the carbon metrics of the funds.

        Returns:
            dict with carbon metrics

        Examples:
            >>> Funds("myria").carbonMetrics()

        """

        return self.GetData("esg/carbonMetrics").json()

    def costIllustration(self) -> dict:
        """
        This function retrieves the cost of the funds.

        Returns:
            dict cost of funds

        Examples:
            >>> Funds("FOUSA00E5P").costIllustration()

        """
        return self.GetData("price/costIllustration").json()
    
    def costProjection(self) -> dict:
        """
        This function retrieves performance with the cost projection.

        Returns:
            dict performance with cost projection

        Examples:
            >>> Funds("FOUSA00E5P").costProjection()

        """
        return self.GetData("price/costProjection").json()

    def couponRange(self)  :
        """
        This function retrieves the coupon of the funds, index and category.

        Returns:
            dict coupon

        Examples:
            >>> Funds("myria").couponRange()

        """
        return self.GetData("process/couponRange").json()

    def creditQuality(self) -> dict:
        """
        This function retrieves the credit notation of the funds, index and category.

        Returns:
            dict credit notation

        Examples:
            >>> Funds("myria").creditQuality()

        """
        return self.GetData("portfolio/creditQuality").json()

    def distribution(self, 
                     period:str="annual") -> dict:
        """
        This function retrieves the coupon distributed by the funds.

        Args:
            period (str) : annual or latest

        Raises:
            ValueError whenever the pariod parameter is not annual or latest

        Returns:
            dict distribution of coupon

        Example:
            >>> Funds("myria").distribution("annual")

        """

        period_choice = ["annual", "latest"]
        if period not in period_choice:
            raise ValueError(
                f"""period parameter can only take one of
                             the values: {", ".join(period_choice)}"""
            )

        return self.GetData(f"distribution/{period}").json()

    def downloadDocument(self,
                         marketId:str,
                         documentType:str,
                         languageId:str,
                         folderPath:str=".") -> dict:
        
        """
        This function download documents.

        Args:
            marketId (str) : country available for the document, represented using two-letter codes
            documentType (str) : document to download
            languageId (str) : languages available for the document, represented using two-letter codes
            folderPath (str) : folder path where to save the file 
        Raises:
            TypeError raised whenever the fund is not available in the marketId

        Returns:
            dict with documents information

        Examples:
            >>> Funds("myria").getDocumentInformation("fr")

        """

        if not isinstance(marketId, str):
            raise TypeError("marketId parameter should be a string")
        
        if not isinstance(documentType, str):
            raise TypeError("marketId parameter should be a string")
        
        if not isinstance(languageId, str):
            raise TypeError("marketId parameter should be a string")

        docInfo = self.getDocumentInformation(marketId)

        if "documents" not in docInfo["components"]:
            raise FileNotFoundError(f"There are no documents available for the {self.asset_type} {self.name} ({self.code})") 
        docFound = False
        docType = []
        foundLanguage = False
        languageAvailable = []
        if docInfo["components"]["documents"]["status"] != 200:
            raise FileNotFoundError(f"There are no documents available for the {self.asset_type} {self.name} ({self.code}) on market {marketId}") 

        for document in docInfo["components"]["documents"]["payload"]:
            docType.append(document["name"])
            if document["name"] == documentType:
                docFound = True
                docId = document["id"]
                for documents in document["documents"]:
                    docExtension =documents["mimeType"]
                    docDate = documents["effectiveDate"]
                    for language in documents["languages"]:
                        documentLanguage = language
                        languageAvailable.append(documentLanguage)
                        if language == languageId:
                            foundLanguage = True
                            break
                    if foundLanguage == True:
                        break
                break
                    
        if docFound == False:
            raise FileNotFoundError(f"""The document type {documentType} 
                                  does not exist for the {self.asset_type}
                                    {self.name} ({self.code}).
                                    Documents type available are {', '.join(docType)}""")
        
        if docExtension != "application/pdf":
            raise FileExistsError(f"""The document type {documentType} 
                                  exists for the {self.asset_type}
                                    {self.name} ({self.code}), but it's extension is {docExtension}.
                                    This type of file is not yet supported  is not by MStarpy.""")
        if foundLanguage == False:
            warnings.warn(f"""Document language {languageId} is not found, 
                          available language are {', '.join(languageAvailable)}.
                          The document in {languageAvailable[-1]} is downloaded.""")
            
        url = f"https://global.morningstar.com/api/v1/{marketId}/investments/{self.asset_type}s/{self.code}/documents/_document"

        params = { "documentId" : docId,
                  "languageId" : documentLanguage}
        
        headers = {"user-agent": random_user_agent()}

        response = requests.get(
            url, params=params, headers=headers, proxies=self.proxies
        )

        timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        fileName = f"{documentType}-{self.code}-{docDate}-{timestamp}.pdf"
        with open(f"{folderPath}/{fileName}", "wb") as f:
            f.write(response.content)


        return {"status" : response.status_code, 
                "mesage" : "File downloaded", 
                "filename" : fileName,
                "folder" : folderPath}
    def equityStyle(self) -> dict:
        """
        This function retrieves the equity style of the funds and category.

        Returns:
            dict equity style

        Examples:
            >>> Funds("myria").equityStyle()

        """
        return self.GetData("process/stockStyle/v2").json()

    def equityStyleBoxHistory(self) -> dict:
        """
        This function retrieves the equity style history of the funds

        Returns:
            dict equity style history

        Examples:
            >>> Funds("myria").equityStyleBoxHistory()

        """
        return self.GetData("process/equityStyleBoxHistory").json()

    def esgData(self) -> dict:
        """
        This function retrieves ESG data of the funds and category

        Returns:
            dict ESG data

        Examples:
            >>> Funds("myria").esgData()

        """

        return self.GetData("esg/v1").json()
    
    def esgRisk(self) -> dict:
        """
        This function retrieves ESG drisk of the funds and category

        Returns:
            dict ESG risk

        Examples:
            >>> Funds("myria").esgRisk()

        """

        return self.GetData("esgRisk").json()

    def factorProfile(self) -> dict:
        """
        This function retrieves the factor profile of the funds, index and category

        Returns:
            dict factor profile

        Examples:
            >>> Funds("myria").factorProfile()

        """
        return self.GetData("factorProfile").json()

    def feeLevel(self)  :
        """
        This function retrieves the fees of the fund compare to its category.

        Returns:
            dict fees

        Examples:
            >>> Funds("myria").feeLevel()

        """
        return self.GetData("price/feeLevel/v1").json()

    def feeMifid(self, 
                 currency:str="EUR") -> dict:
        """
        This function retrieves the fees of the fund.

        Returns:
            dict fees

        Examples:
            >>> Funds("myria").feeMifid()

        """
        return self.ltData("Mifid", currency=currency)

    def financialMetrics(self) -> dict:
        """
        This function retrieves the final metrics of the funds and category.

        Returns:
            dict financial metrics

        Examples:
            >>> Funds("myria").financialMetrics()

        """
        return self.GetData("process/financialMetrics").json()

    def fixedIncomeStyle(self) -> dict:
        """
        This function retrieves the fixed income style of the funds and category.

        Returns:
            dict fixed income style

        Examples:
            >>> Funds("myria").fixedIncomeStyle()

        """

        return self.GetData("process/fixedIncomeStyle").json()

    def fixedincomeStyleBoxHistory(self) -> dict:
        """
        This function retrieves the fixed income style history of the funds.

        Returns:
            dict fixed income style

        Examples:
            >>> Funds("myria").fixedincomeStyleBoxHistory()

        """
        return self.GetData("process/fixedincomeStyleBoxHistory").json()

    def graphData(self) -> dict:
        """
        This function retrieves historical Total Assets and Net Flow of the funds.

        Returns:
            dict historical data

        Examples:
            >>> Funds("myria").graphData()

        """

        return self.GetData("parent/graphData").json()

    def getDocumentInformation(self, 
                    marketId:str
                    ) -> dict:
        """
        This function retrieves information about documents.
        Args:
            marketId (str) : country in two letters from where the document is available

        Raises:
            TypeError raised whenever the fund is not available in the marketId

        Returns:
            dict with documents information

        Examples:
            >>> Funds("myria").getDocumentInformation("fr")

        """

        if not isinstance(marketId, str):
            raise TypeError("marketId parameter should be a string")

        
        # url of API
        url = f"""https://global.morningstar.com/api/v1/{marketId}/investments/{self.asset_type}s/{self.code}/documents"""

        params = { "marketId" : marketId}
        headers = {"user-agent": random_user_agent()}

        response = requests.get(
            url, params=params, headers=headers, proxies=self.proxies
        )

        response_json = response.json()
        print(response_json)
        if "message" in response_json and response.status_code == 404:
            if response_json["message"] == 'Security Market Access Error':
                raise ValueError(f"marketId paramater can only take one of these values {', '.join(response_json['allowedMarketIds'])} ")

        return response_json

    def historicalData(self, 
                       version:int=4) -> dict:
        """
        This function retrieves the historical price of the funds, index and category

        Args:
            version (int) : version of the api of historical data from 2 to 5
        
        Returns:
            dict with historical data

        Examples:
            >>> Funds("myria").historicalData()

        """
        if not isinstance(version,int):
            raise TypeError("version paramater should be an integer")
        
        if version not in range(2,5):
            raise ValueError("version paramater should be between 2 and 5")

        return self.GetData(f"performance/v{version}", url_suffix="").json()

    def historicalExpenses(self) -> dict:
        """
        This function retrieves historical expenses of the funds.

        Returns:
            dict historical expenses

        Examples:
            >>> Funds("myria").historicalExpenses()

        """
        if self.asset_type == "etf":
            return {}
        return self.GetData("price/historicalExpenses").json()
    
    def historicalRating(self) -> dict:
        """
        This function retrieves MorningStar historical rating of the fund.

        Returns:
            dict historical rating

        Examples:
            >>> Funds("myria").historicalRating()

        """

        return self.GetData("morningstarTake/historicalRating").json()

    def holdings(self, 
                 holdingType: str = "all") -> pd.DataFrame:
        """
        This function retrieves holdings of the funds.

        Args:
            holdingType (str) : paramater to select the kind of holdings; all, bond, equity or other

        Returns:
            pandas DataFrame holdings

        Raises:
            ValueError whenever the parameter is not all, bond, equity or other

        Examples:
            >>> Funds("myria").holdings("all")
            >>> Funds("myria").holdings("bond")
            >>> Funds("myria").holdings("equity")
            >>> Funds("myria").holdings("other")

        """
        holdingType_to_holdingPage = {
            "all": "all",
            "bond": "boldHoldingPage",
            "equity": "equityHoldingPage",
            "other": "otherHoldingPage",
        }
        if holdingType not in holdingType_to_holdingPage:
            raise ValueError(
                f"""parameter holdingType must take one of the following values
                : {", ".join(holdingType_to_holdingPage.keys())}"""
            )

        if holdingType == "all":
            return pd.DataFrame(
                self.position()["equityHoldingPage"]["holdingList"]
                + self.position()["boldHoldingPage"]["holdingList"]
                + self.position()["otherHoldingPage"]["holdingList"]
            )
        else:
            return pd.DataFrame(
                self.position()[holdingType_to_holdingPage[holdingType]]["holdingList"]
            )

    def investmentFee(self) -> dict:
        """
        This function retrieves the investment fees.

        Returns:
            dict investment fees

        Examples:
            >>> Funds("LU0823421689").investmentFee()

        """
        return self.GetData("price/investmentFee").json()
    
    def investmentLookup(self, 
                         currency:str="EUR") -> dict:
        """
        This function gives details about fund investment.

        Returns:
            dict fund investment

        Examples:
            >>> Funds("myria").investmentLookup()

        """
        return self.ltData("investmentTypeLookup", currency=currency)

    def investmentStrategy(self) -> dict:
        """
        This function retrieves the investment strategy.

        Returns:
            dict investment strategy

        Examples:
            >>> Funds("LU0823421689").investmentStrategy()

        """
        return self.GetData("morningstarTake/investmentStrategy").json()

    def marketCapitalization(self) -> dict:
        """
        This function retrieves the marketCapitalization breakdown of the funds,
        category and index.

        Returns:
            dict market capitalization

        Examples:
            >>> Funds("myria").marketCapitalization()

        """
        return self.GetData("process/marketCap").json()

    def maturitySchedule(self) -> dict:
        """
        This function retrieves the maturity breakdown of the funds and category.

        Returns:
            dict maturity

        Examples:
            >>> Funds("myria").maturitySchedule()

        """
        return self.GetData("process/maturitySchedule").json()

    def maxDrawDown(self, 
                    year:int=3) -> dict:
        """
        This function retrieves the max drawdown of the funds, index and category.

        Args:
            year (int) : period of calculation in year

        Returns:
            dict max drawdown

        Raises:
            TypeError whenever the parameter year is not an integer

        Examples:
            >>> Funds("myria").maxDrawDown()

        """

        if not isinstance(year, int):
            raise TypeError("year parameter should be an integer")

        return self.GetData(
            "performance/marketVolatilityMeasure", params={"year": year}
        ).json()

    def medaListComparables(self) -> dict:
        """
        This function retrieves comparable funds.

        Returns:
            dict comparable funds

        Examples:
            >>> Funds("myria").medaListComparables()

        """

        return self.GetData("medaListComparables").json()
    


    def metaData(self) -> dict:
        """
        This function retrieves meta Data about the funds.

        Returns:
            dict meta Data

        Examples:
            >>> Funds("myria").metaData()

        """

        return self.GetData("securityMetaData",url_suffix="").json()
    
    def morningstarAnalyst(self) -> dict:
        """
        This function retrieves the raiting of MorningStar analyst.

        Returns:
            dict rating

        Examples:
            >>> Funds("myria").morningstarAnalyst()

        """

        return self.GetData("morningstarAnalyst").json()
    
    def morningstarOpinion(self,
                           version:int=3) -> dict:
        """
        This function retrieves the opinion of MorningStar analyst.

        Args:
            version (int) : version of the api of morningstarOpinion shoulb be 2 or 3

        Returns:
            dict opinion

        Examples:
            >>> Funds("myria").morningstarOpinion()

        """

        if not isinstance(version,int):
            raise TypeError("version paramater should be an integer")
        
        if version not in range(2,4):
            raise ValueError("version paramater should be 2 or 3")

        return self.GetData(f"morningstarTake/v{version}",url_suffix="").json()

    def multiLevelFixedIncomeData(self, 
                                  primary:str="superEffectiveDuration", 
                                  secondary:str="superSector.weight") -> dict:
        """
        This function retrieves the exposures of fixed income

        Args:
            primary (str) : can be superEffectiveDuration, superYieldToWorst, creditQuality
            secondary (str) : can be superSector.weight, region.weight, creditQuality.weight

        Returns:
            dict exposures

        Raises:
            ValueError whenever the primary and seconday paramater is not a value expected

        Examples:
            >>> Funds("myria").multiLevelFixedIncomeData()

        """

        primary_choice = [
            "superEffectiveDuration",
            "superYieldToWorst",
            "creditQuality",
        ]
        secondary_choice = [
            "superSector.weight",
            "region.weight",
            "creditQuality.weight",
        ]

        if primary not in primary_choice:
            raise ValueError(
                f"""primary parameter can only take one of the 
                values : {", ".join(primary_choice)}"""
            )

        if secondary not in secondary_choice:
            raise ValueError(
                f"""secondary parameter can only take one of the values 
                : {", ".join(secondary_choice)}"""
            )

        if primary == "creditQuality" and secondary == "creditQuality.weight":
            raise ValueError(
                f"primary and secondary parameters cannot be both credit quality"
            )

        return self.GetData(
            "multiLevelFixedIncomeData",
            params={"primary": primary, "secondary": secondary},
        ).json()

    def nav(self, 
            start_date:datetime.datetime,
            end_date: datetime.datetime, 
            frequency:str="daily") -> list[dict]:
        """
        This function retrieves the NAV of the funds

        Args:
            start_date (datetime) : start date to get nav
            end_date (datetime) : end date to get nav
            frequency (str) : can be daily, weekly, monthly

        Returns:
            list of dict with nav

            >>> Funds("myria").nav(datetime.datetime.today() 
            - datetime.timedelta(30),datetime.datetime.today())

        Raises:
            TypeError: raised whenever the parameter type is not the type expected
            ValueError : raised whenever the parameter is not valid or no funds found

        """

        return self.TimeSeries(
            ["nav", "totalReturn"],
            start_date=start_date,
            end_date=end_date,
            frequency=frequency,
        )

    def otherFee(self) -> dict:
        """
        This function retrieves the other fee of the etf

        Returns:
            dict fees

        Examples:
            >>> Funds("American Century Foc Dynmc Gr ETF").otherFee()

            {'expenseWaiver': False, 'expenseReimbursement': None, 
            'expirationDate': None, 'expenseWaivers': None}

        """
        return self.GetData("price/otherFee").json()

    def ownershipZone(self) -> dict:
        """
        This function retrieves ownershipZone of the funds, index and category.

        Returns:
            dict ownershipZone

        Examples:
            >>> Funds("myria").ownershipZone()

        """

        return self.GetData("process/ownershipZone").json()

    def parentMedal(self) -> list[dict]:
        """
        This function retrieves medal funds from the asset manager.

        Returns:
            list of dict rating parent medal funds

        Examples:
            >>> Funds("myria").parentMedal()

        """

        return self.GetData("parent/medalistRating/topfunds").json()
    
    def parentMedaListRating(self) -> list[dict]:
        """
        This function retrieves summary medal funds from the asset manager.

        Returns:
            list of dict rating parent funds

        Examples:
            >>> Funds("myria").parentmedaListRating()

        """

        return self.GetData("parent/medalistRating").json()
    
    def parentMstarRating(self,
                          ) -> list[dict]:
        """
        This function retrieves the rating of asset manager funds.

        Returns:
            list of dict rating

        Examples:
            >>> Funds("myria").parentMstarRating()

        """

        return self.GetData("parent/parentMstarRating").json()
    
    def parentRatingRecentChange(self) -> list[dict]:
        """
        This function retrieves recent change in funds rating from the asset manager.

        Returns:
            list of dict rating parent funds

        Examples:
            >>> Funds("myria").parentRatingRecentChange()

        """

        return self.GetData("parent/medalistRating/topfundsUpDown").json()

    def parentSummary(self) -> dict:
        """
        This function retrieves info about the parent.

        Returns:
            dict parent info

        Examples:
            >>> Funds("myria").parentSummary()

        """
        return self.GetData("parent/parentSummary").json()

    def people(self) -> dict:
        """
        This function retrieves info about people who works in the company.

        Returns:
            dict people info

        Examples:
            >>> Funds("myria").people()

        """
        return self.GetData("people").json()
    
    def performanceTable(self) -> dict:
        """
        This function retrieves the performance of fund, catgory and index

        Returns:
            dict performance

        Examples:
            >>> Funds("myria").performanceTable()

        """
        return self.GetData("performance/table", url_suffix="").json()
    
    def position(self) -> dict:
        """
        This function retrieves the hodings of the funds.

        Returns:
            dict holdings

        Examples:
            >>> Funds("myria").position()

        """

        return self.GetData(
            "portfolio/holding/v2", params={"premiumNum": 10000, "freeNum": 10000}
        ).json()

    def proxyVotingManagement(self) :
        """
        This function retrieves the vote of management.

        Returns:
            dict vote

        Examples:
            >>> Funds("myria").proxyVotingManagement()

        """
        return self.GetData("people/proxyVoting/management").json()

    def proxyVotingShareHolder(self) -> dict:
        """
        This function retrieves the vote of shareholders.

        Returns:
            dict vote

        Examples:
            >>> Funds("myria").proxyVotingShareHolder()

        """
        return self.GetData("people/proxyVoting/shareHolder").json()

    def productInvolvement(self) -> dict:
        """
        This function retrieves the involvement of the funds

        Returns:
            dict involvement

        Examples:
            >>> Funds("myria").proxyVotingShareHolder()

        """

        return self.GetData("esg/productInvolvement").json()



    def quote(self, 
                version:int=7) -> dict:
        """
        This function retrieves general information about the funds.

        Args:
            version (int) : version of the api of historical data from 2 to 7
        
        Returns:
            dict with general information

        Examples:
            >>> Funds("myria").quote()

        """
        if not isinstance(version,int):
            raise TypeError("version paramater should be an integer")
        
        if version not in range(1,8):
            raise ValueError("version paramater should be between 2 and 7")

        return self.GetData(f"quote/v{version}").json()
    
    def regionalSector(self) -> dict:
        """
        This function retrieves the breakdown of the funds, category and index by region

        Returns:
            dict regional breakdown

        Examples:
            >>> Funds("myria").regionalSector()

        """
        return self.GetData("portfolio/regionalSector").json()

    def regionalSectorIncludeCountries(self) -> dict:
        """
        This function retrieves the breakdown of the funds,
        category and index by region and country

        Returns:
            dict regional breakdown

        Examples:
            >>> Funds("myria").regionalSectorIncludeCountries()

        """
        return self.GetData("portfolio/regionalSectorIncludeCountries").json()

    def riskReturnScatterplot(self) -> dict:
        """
        This function retrieves the return and standard
        deviation of the funds and category

        Returns:
            dict risk return

        Examples:
            >>> Funds("myria").riskReturnScatterplot()

        """
        return self.GetData("performance/riskReturnScatterplot").json()

    def riskReturnSummary(self) -> dict:
        """
        This function retrieves the return and risk summary 
        of the funds compare to the category

        Returns:
            dict risk return

        Examples:
            >>> Funds("myria").riskReturnSummary()

        """

        return self.GetData("performance/riskReturnSummary").json()
    
    def riskScore(self) -> dict:
        """
        This function retrieves the risk score of the fund.

        Returns:
            dict risk Score

        Examples:
            >>> Funds("myria").riskScore()

        """
        return self.GetData("performance/riskScore").json()

    def riskVolatility(self) -> dict:
        """
        This function retrieves the alpha, beta, R², 
        volatility and Sharpe ratio of the funds, category and index.

        Returns:
            dict econometrics

        Examples:
            >>> Funds("myria").riskVolatility()

        """
        return self.GetData("performance/riskVolatility").json()

    def salesFees(self) -> dict:
        """
        This function retrieves the sales fees of the funds

        Returns:
            dict fees

        Examples:
            >>> Funds("myria").salesFees()

        """
        if self.asset_type == "etf":
            return {}
        return self.GetData("price/salesFees").json()


    def sector(self, 
                version:int=2) -> dict:
        """
        This function retrieves the sector breakdown of the funds, category and index
        Args:
            version (int) : version of the api of allocation map, 1 or 2

        Returns:
            dict sector breakdown

        Examples:
            >>> Funds("myria").sector()

        """
        if not isinstance(version,int):
            raise TypeError("version paramater should be an integer")
        
        if version not in range(1,3):
            raise ValueError("version paramater should be 1 or 2")

        return self.GetData(f"portfolio/v{version}/sector").json()

    def snapshot(self, 
                 currency:str="EUR"):
        """
        This function returns a snapshot of the fund and asset manager.

        Returns:
            dict snapshot fund and asset manager

        Examples:
            >>> Funds("myria").snapshot()

        """
        return self.ltData("MFsnapshot", currency=currency)

    def starRatingFundAsc(self) -> dict:
        """
        This function retrieves the MorningStar rating of the funds
        of the company by ascending order

        Returns:
            dict rating

        Examples:
            >>> Funds("myria").starRatingFundAsc()

        """

        return self.GetData("parent/mstarRating/StarRatingFundAsc").json()

    def starRatingFundDesc(self) -> dict:
        """
        This function retrieves the MorningStar rating of the funds
        of the company by descending order

        Returns:
            dict rating

        Examples:
            >>> Funds("myria").starRatingFundDesc()

        """

        return self.GetData("parent/mstarRating/StarRatingFundDesc").json()


    def strategyPreview(self) -> dict:
        """
        This function retrieves general information and return on funds

        Returns:
            dict returns and general information

        Examples:
            >>> Funds("myria").strategyPreview()

        """
        return self.GetData("strategyPreview").json()
    
    def sustainability(self, 
                       currency:str="EUR") -> dict:
        """
        This function retrieves the sustainability data of the fund.

        Returns:
            dict sustanability data

        Examples:
            >>> Funds("myria").sustainability()

        """
        return self.ltData("sustainability", currency=currency)

    def taxes(self) -> dict:
        """
        This function retrieves the other fee of the etf

        Returns:
            dict taxes

        Examples:
            >>> Funds("American Century Foc Dynmc Gr ETF").taxes()

        """
        return self.GetData("price/taxes").json()


    def trailingReturn(self, 
                       duration:str="daily", 
                       version:int=3) -> dict:
        """
                This function retrieves the trailing return of the funds of the company.

        Args:
        duration (str) : frequency of return can be daily, monthly or quarterly
        version (int) : version of the api of historical data from 1 to 3

        Returns:
            dict trailing return

        Raises:
            ValueError whenever the parameter duration is not daily, monthly or quarterly

        Example:
            >>> Funds("myria").trailingReturn("daily")

        """

        if not isinstance(duration,str):
            raise TypeError("duration paramater should be a string")
        
        if not isinstance(version,int):
            raise TypeError("version paramater should be an integer")
        
        if version not in range(1,4):
            raise ValueError("version paramater should be between 1 and 3")
        
        duration_choice = ["daily", "monthly", "quarterly"]
        if duration not in duration_choice:
            raise ValueError(
                f'duration parameter can only take one of the values: {", ".join(duration_choice)}'
            )

        return self.GetData(f"trailingReturn/v{version}", {"duration": duration}).json()

